import argparse
import json
import os
import time
from pathlib import Path

import numpy as np
import openai

from util.apidefs import get_api_defs

api_defs = None


def write_code(code, outdir):
    if not os.path.exists(outdir):
        os.makedirs(outdir, exist_ok=True)
    i = 1
    outdir = Path(outdir)
    while True:
        if os.path.exists(outdir / f"{i}.py"):
            i += 1
            continue
        with open(outdir / f"{i}.py", "w") as fw:
            fw.writelines(code)
        return


def gen_prompt(prompt_mode, prompt_template, api, lib):
    # TODO: do special transformation for tf.raw_ops
    # For example, tf.raw_ops.Abs(x,name=None) must be called with
    # tf.raw_ops.Abs(x=x_,name=name_)
    lib_str = "PyTorch" if lib == "torch" else "TensorFlow 2.10.0"
    if prompt_mode == "docstring_steps":
        """
        ...
        Task 3: Call the API tf.math.accumulate_n(inputs,shape=None,tensor_dtype=None,name=None)
        """
        api_def = api
        if api in api_defs:
            api_def = api_defs[api]["def"]
        prompt = prompt_template.format(lib_str, api_def)
    else:
        """
        ...
        Task 3: Call the API tf.math.accumulate_n
        """
        prompt = prompt_template.format(lib_str, api)
    return prompt


def gen(
    engine,
    prompt,
    temperature,
    max_tokens,
    n_samples,
    top_p,
    frequency_penalty,
    presence_penalty,
):
    t_start = time.time()
    response = openai.Completion.create(
        engine=engine,
        prompt=prompt,
        temperature=temperature,
        max_tokens=max_tokens,
        top_p=top_p,
        n=n_samples,
        frequency_penalty=frequency_penalty,
        presence_penalty=presence_penalty,
        # stop=["#"]
    )
    t_end = time.time()
    codes = [prompt + x["text"] for x in response["choices"]]
    return codes, t_end - t_start


def edit(input, instruction):
    response = openai.Edit.create(
        engine="code-davinci-edit-001",
        input=input,
        instruction=instruction,
        temperature=0,
        top_p=1,
    )
    return response["choices"][0]["text"]


if __name__ == "__main__":
    parser = argparse.ArgumentParser()

    parser.add_argument(
        "--mode", type=str, default="completion", help="one of 'completion', 'edit' "
    )
    parser.add_argument("--lib", type=str, default="tf", help="one of 'tf', 'torch'")
    parser.add_argument(
        "--outdir", type=str, default="../codex_seed_programs_100sample"
    )
    parser.add_argument(
        "--prompt_template",
        type=str,
        default="data/codex_prompt/docstring_steps.txt",
    )
    parser.add_argument(
        "--input", type=str, default="data/tf_apis_100sample.txt", help="list of apis"
    )

    # Codex API parameters
    parser.add_argument(
        "--engine", type=str, default="code-davinci-002", help="Codex engine name"
    )
    parser.add_argument("--apikey", type=int, default=0, help="Codex API key")
    parser.add_argument("--temperature", type=float, default=0.4)
    parser.add_argument("--n_samples", type=int, default=25)
    parser.add_argument("--max_tokens", type=int, default=256)
    parser.add_argument("--top_p", type=float, default=0.95)
    parser.add_argument("--frequency_penalty", type=float, default=0.0)
    parser.add_argument("--presence_penalty", type=float, default=0.0)

    args = parser.parse_args()

    mode = args.mode
    lib = args.lib
    temperature = args.temperature
    n_samples = args.n_samples
    max_tokens = args.max_tokens
    with open(args.prompt_template, "r") as f:
        prompt_template = f.read()

    api_keys = {
        0: "YOUR_API_KEY",
    }
    openai.api_key = api_keys[args.apikey]

    version = "raw"
    prompt_mode = args.prompt_template.split("/")[-1].rsplit(".", 1)[0]
    code_dir = (
        Path(args.outdir)
        / prompt_mode
        / f"codex_{lib}_{temperature}_{max_tokens}_{n_samples}"
        / f"{version}"
    )
    print("Task: ", code_dir)
    os.makedirs(code_dir, exist_ok=True)
    with open(code_dir / "args.txt", "w") as f:
        f.write(str(args))

    # Load api defs for some prompt templates to use
    api_defs = get_api_defs(lib)
    max_patience = 5
    num_limit_reached = 0

    if mode == "completion":
        if lib == "tf" or lib == "torch":
            wait_seconds = 0.0
            with open(args.input, "r") as fr:
                apis = fr.readlines()
                apis = [a.strip() for a in apis]
            out_dir = code_dir
            tcosts = []
            flag = True
            ret = {}
            apicnt = 0
            while flag:
                err_flag = False
                suc_flag = False

                for api in apis:
                    if os.path.exists(out_dir / api):
                        continue
                    if wait_seconds > 0.0:
                        time.sleep(wait_seconds)
                    print(api)
                    prompt = gen_prompt(prompt_mode, prompt_template, api, lib)
                    try:
                        codes, tc = gen(
                            args.engine,
                            prompt,
                            temperature,
                            max_tokens,
                            n_samples,
                            args.top_p,
                            args.frequency_penalty,
                            args.presence_penalty,
                        )
                        tcosts.append(tc)
                        ret[api] = {"code": codes, "g_time": tc}
                        for code in codes:
                            write_code(code, out_dir / api)
                        suc_flag = True
                        apicnt += 1
                    except Exception as e:
                        err_flag = True
                        if wait_seconds < 3.0:
                            if num_limit_reached >= max_patience:
                                wait_seconds += 0.5
                                num_limit_reached = 0
                            time.sleep(30.0)
                        else:
                            time.sleep(30.0)
                if err_flag:
                    flag = True
                if not suc_flag:
                    flag = False
            with open(out_dir / "outputs.json", "w") as f:
                json.dump(ret, f)
            with open(out_dir / "log.txt", "w") as f:
                f.write(
                    "{}\nAverage cost in seconds for each Codex API call".format(
                        np.average(tcosts)
                    )
                )
